
/* leds-dspwm.c: Sony DisplayService-PWM LED driver.
 *
 * Copyright 2019 Sony Corporation.
 * Author: Sony Corporation.
 *
 * This package is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <linux/module.h>
#include <linux/device.h>
#include <linux/sysfs.h>
#include <linux/slab.h>
#include <linux/pagemap.h>
#include <linux/leds.h>
#include <linux/platform_device.h>
#include <mach/mt_pwm.h>
#include <leds-dspwm.h>

#define DRV_NAME				"leds-dspwm"

#define DBGTRACE(fmt, ...)			pr_debug("\e[32m[" DRV_NAME "] "       fmt "\e[0m\n", ## __VA_ARGS__)
#define ERRTRACE(fmt, ...)			pr_err("\e[31m[" DRV_NAME "] error " fmt "\e[0m\n", ## __VA_ARGS__)

#define BRIGHTNESS_BITS                         (8)

#define BRIGHTNESS_MASK                         ((1 << BRIGHTNESS_BITS) - 1)

#define LED_BRIGHTNESS_MAX                      (0xff)
#define PWM_DUTY_MAX                            (100)

#define PWM_DEFAULT_OUTPUT_FREQ                 (200)

#define PWM_BASE_FREQ                           (32768)

/* LED */
struct led {
	struct led_classdev					cdev;							/* LED Device Class */
	u8							count;							/* Port Number      */
	u8							dummy;							/* Dummy            */
	s16							pwm_port[4];						/* PWM Port Number  */
};

/* private data */
struct dspwm_priv {
	int							count;							/* LED Number */
	struct led						led[1];							/* LED */
};


/* -------------------------------------------------------------------------------- */
/*   initialize_pwm_config                                                          */
/*                                                                                  */
/*   IN : pwm_easy_config = MTK PWM Easy Config                                     */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int initialize_pwm_config(struct pwm_easy_config *p_conf)
{
	DBGTRACE("******** %s() ********\n", __func__);

	/* Initialize pwm config */
	p_conf->pwm_no                        = PWM_MIN;
	p_conf->duration                      = (PWM_BASE_FREQ /
						PWM_DEFAULT_OUTPUT_FREQ) - 1;
	p_conf->duty                          = 0;
	p_conf->clk_div                       = CLK_DIV1;
	p_conf->clk_src                       = PWM_CLK_OLD_MODE_32K;
	p_conf->pmic_pad                      = false;
	return 0;
}

/* -------------------------------------------------------------------------------- */
/*   Set PWM Configuration.                                                         */
/*                                                                                  */
/*   IN : pwm_port = PWM Port No (PWM1-PWM5)                                        */
/*        value  = LED Brightness (0-0xff)                                          */
/*   OUT: 0    == Normal                                                            */
/*        0    != Error                                                             */
/* -------------------------------------------------------------------------------- */
static int set_pwm_conf(u16 pwm_port, uint8_t value)
{
	int			ret;
	struct pwm_easy_config	pwm_conf;
	unsigned short		pwm_duty_cycle;

	DBGTRACE("%s(%d) pwm_port:%u value:%u", __func__, __LINE__, pwm_port, value);

	initialize_pwm_config(&pwm_conf);

	pwm_conf.pwm_no = pwm_port;

	pwm_duty_cycle = ((value * PWM_DUTY_MAX) + LED_BRIGHTNESS_MAX - 1) /
			 LED_BRIGHTNESS_MAX;

	pwm_conf.duty = pwm_duty_cycle;

	DBGTRACE("%s(%d) duty:%d duration:%d", __func__, __LINE__, pwm_conf.duty,
		 pwm_conf.duration);
	ret = pwm_set_easy_config(&pwm_conf);
	if (ret < 0) {
		ERRTRACE("%s()[%d] : Call pwm_set_spec_config() return %d\n", __func__,
			 __LINE__, ret);
		return -EFAULT;
	}
	return 0;
}

/* -------------------------------------------------------------------------------- */
/*   Set LED Brightness.                                                            */
/*                                                                                  */
/*   IN : p_cdev = LED Device Class                                                 */
/*        value  = LED Brightness                                                   */
/*   OUT: none                                                                      */
/* -------------------------------------------------------------------------------- */
static void set_brightness(struct led_classdev *p_cdev,
			   enum led_brightness value)
{
	struct led		*p_led = container_of(p_cdev, struct led, cdev);
	int			index;

	for (index = 0; index < p_led->count; index++) {
		if (p_led->pwm_port[index] >= PWM_MIN && p_led->pwm_port[index] < PWM_MAX) {
			set_pwm_conf(p_led->pwm_port[index],
				     (value >> (index * BRIGHTNESS_BITS))&BRIGHTNESS_MASK);
		}
	}
}

/* -------------------------------------------------------------------------------- */
/*   Set LED.                                                                       */
/*                                                                                  */
/*   IN : p_led  = LED                                                              */
/*        p_data = LED Data                                                         */
/*   OUT: none                                                                      */
/* -------------------------------------------------------------------------------- */
static void set_led(struct led *p_led, struct ledsdspwmdata *p_data)
{
	s16		port;
	int		index;
	u32		max = 0x00000000;

	/* Initialize */
	p_led->cdev.name           = &p_data->p_name[0];
	p_led->cdev.brightness     = LED_OFF;
	p_led->cdev.brightness_set = set_brightness;
	p_led->count               = p_data->count;

	for (index = 0; index < p_led->count; index++) {

		/* Port No. */
		port = p_data->port[index];
		p_led->pwm_port[index] = port;

		/* Get Brightness Max */
		if (p_led->pwm_port[index] >= PWM_MIN && p_led->pwm_port[index] < PWM_MAX)
			max |= (BRIGHTNESS_MASK << (index * BRIGHTNESS_BITS));
		else
			max |= (0x00 << (index * BRIGHTNESS_BITS));
	}

	p_led->cdev.max_brightness = max;
}

/* -------------------------------------------------------------------------------- */
/*   Probe.                                                                         */
/*                                                                                  */
/*   IN : p_dev = Platform Device.                                                  */
/*   OUT: 0    == Normal                                                            */
/*        0    != Error                                                             */
/* -------------------------------------------------------------------------------- */
static int __init ledsdspwm_probe(struct platform_device *p_dev)
{
	struct dspwm_priv			*p_priv;
	struct ledsdspwmdata			*p_data;
	struct led				*p_led;
	int					result, index, count;

	DBGTRACE("%s(%d)", __func__, __LINE__);
	/* Get LED Data Num */
	count  = 0;
	p_data = dev_get_platdata(&p_dev->dev);
	while (p_data->p_name != NULL) {
		count++;
		p_data++;
	}

	/* allocate private data */
	p_priv = kzalloc(offsetof(struct dspwm_priv,
				  led) + (sizeof(struct led) * count), GFP_KERNEL);
	if (p_priv == NULL)
		return -ENOMEM;

	p_priv->count = count;

	/* register LED Class Device */
	p_led  = &p_priv->led[0];
	p_data = dev_get_platdata(&p_dev->dev);
	for (index = 0; index < count; index++, p_led++, p_data++) {
		set_led(p_led, p_data);
		result = led_classdev_register(NULL, &p_led->cdev);
		if (!result) {
			result = sysfs_chmod_file(&p_led->cdev.dev->kobj,
						  &p_led->cdev.dev->class->dev_attrs->attr, 0666);
		}
		if (result < 0) {
			ERRTRACE("%s(%d)", __func__, __LINE__);
			for (--index; index >= 0; index--)
				led_classdev_unregister(&p_priv->led[index].cdev);

			goto ledsdspwm_probe_f00;
		}
	}

	platform_set_drvdata(p_dev, p_priv);

	return 0;

ledsdspwm_probe_f00:
	kfree(p_priv);

	return result;
}

/* -------------------------------------------------------------------------------- */
/*   Remove                                                                         */
/*                                                                                  */
/*   OUT: 0 == Normal                                                               */
/*        0 != Error                                                                */
/* -------------------------------------------------------------------------------- */
static int ledsdspwm_remove(struct platform_device *p_dev)
{
	struct dspwm_priv		*p_priv = platform_get_drvdata(p_dev);
	int				index;


	for (index = 0; index < p_priv->count; index++)
		led_classdev_unregister(&p_priv->led[index].cdev);

	kfree(p_priv);

	return 0;
}


/* -------------------------------------------------------------------------------- */
/*   Resume                                                                         */
/*                                                                                  */
/*   IN : p_dev = Platform Device                                                   */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int ledsdspwm_resume(struct platform_device *p_dev)
{
	struct dspwm_priv		*p_priv = platform_get_drvdata(p_dev);
	struct led			*p_led;
	int				index;


	DBGTRACE("%s(%d)", __func__, __LINE__);

	/* Get Brightness */
	p_led  = &p_priv->led[0];
	for (index = 0; index < p_priv->count; index++, p_led++) {
		p_led->cdev.brightness = LED_OFF;
	}

	return 0;
}

static struct platform_driver		ledsdspwm_platform_driver = {
	.probe   = ledsdspwm_probe,
	.remove  = ledsdspwm_remove,
	.resume  = ledsdspwm_resume,
	.driver  = {
		.name  = DRV_NAME,
		.owner = THIS_MODULE,
	},
};

module_platform_driver(ledsdspwm_platform_driver);

MODULE_DESCRIPTION("Sony DS-PWM LED driver");
MODULE_AUTHOR("Sony Corporation");
MODULE_LICENSE("GPL");
MODULE_ALIAS("Sony DS-PWM LED driver");
MODULE_VERSION("v1.00");
